//-----------------------------------------------------------------------------
// Copyright (c) 2006 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

#include <cassert>
#include <iostream>

#if defined(_WIN32)
#include <crtdbg.h>
#endif

#include "SmartPtr.h"

//-----------------------------------------------------------------------------

class Mammal : public Referable
{
public:
    Mammal()
    { std::cout << "Mammal constructor called" << std::endl; }

    ~Mammal()
    { std::cout << "Mammal destructor called" << std::endl; }

    virtual void performAction() const
    { std::cout << "Mammal performing action" << std::endl; }
};

class Dog : public Mammal
{
public:
    Dog()
    { std::cout << "Dog constructor called" << std::endl; }
    
    ~Dog()
    { std::cout << "Dog destructor called" << std::endl; }

    virtual void performAction() const
    { std::cout << "Dog performing action" << std::endl; }
};

class Robot : public Referable
{
public:
    Robot()
    { std::cout << "Robot constructor called" << std::endl; }

    ~Robot()
    { std::cout << "Robot destructor called" << std::endl; }
};

typedef SmartPtr<Mammal> MammalPtr;
typedef SmartPtr<Dog> DogPtr;
typedef SmartPtr<Robot> RobotPtr;

//-----------------------------------------------------------------------------

void Test1()
{
    std::cout << "\tTest 1: Invoking a method from a SmartPtr." << std::endl;

    MammalPtr m = new Dog;

    m->performAction();
}

void Test2()
{
    std::cout << "\n\tTest 2: Ownership testing." << std::endl;

    DogPtr d1, d2;
    Dog *pDog = new Dog;
    
    assert(pDog->refCount() == 0);
    
    d1 = pDog;
    assert(pDog->refCount() == 1);
    
    d2 = d1;
    assert(pDog->refCount() == 2);
    
    d1 = 0;                             // assigning 0 releases ownership
    assert(pDog->refCount() == 1);

    d1 = pDog;
    assert(pDog->refCount() == 2);

    d1 = pDog;
    assert(pDog->refCount() == 2);
}

void Test3()
{
    std::cout << "\n\tTest 3: Equality and inequality testing." << std::endl;

    Dog *pDog = 0;
    DogPtr d1, d2;

    assert(!d1);            // direct test for null pointer
    assert(d1 == 0);        // explicit test for null pointer
    assert(d1 == d2);       // comparison of two smart pointers
    assert(d1 == pDog);     // comparison with a raw pointer

    pDog = new Dog;
    d1 = pDog;
    assert(d1);             // direct test for non-null pointer
}

void Test4()
{
    std::cout << "\n\tTest 4: Inheritance type conversion testing." << std::endl;

    DogPtr d = new Dog;
    MammalPtr m = d;        // ok: since Dog is derived from Mammal
    //RobotPtr r = m;         // error: Robot and Mammal are not related
}

void Test5()
{
    std::cout << "\n\tTest 5: Get raw pointer testing." << std::endl;

    DogPtr d;

    assert(DogPtr::getImpl(d) == 0);
    
    d = new Dog;
    assert(DogPtr::getImpl(d) != 0);
}

int main()
{
    #if defined(_WIN32)
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
    #endif
    
    Test1();
    Test2();
    Test3();
    Test4();
    Test5();

    std::cout << std::endl << "Finished. Press enter to exit.";
    std::cin.get();
 
    return 0;
}